
import requests
import urllib.request
from unittest import TestCase
import datadiligence as dd
from datadiligence.rules import XRobotsTagHeader
import time

# starting local server to echo back headers
from werkzeug.serving import make_server
from server.app import app
import threading


class XRobotsTest(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.server = make_server('localhost', 5001, app)
        cls.server_thread = threading.Thread(target=cls.server.serve_forever)
        cls.server_thread.start()
        time.sleep(1)  # wait for server to start

        cls.rule = XRobotsTagHeader(user_agent="spawningbot")
        cls.rule_2 = XRobotsTagHeader(user_agent=None)

    def test_noheader(self):
        self.assertTrue(self.rule._eval_header_value(""))
        self.assertTrue(self.rule._eval_header_value(None))
        self.assertTrue(self.rule_2._eval_header_value(""))
        self.assertTrue(self.rule_2._eval_header_value(None))

    def test_noai(self):
        self.assertFalse(self.rule._eval_header_value("noai"))
        self.assertFalse(self.rule._eval_header_value("noimageai"))
        self.assertFalse(self.rule._eval_header_value("other, noai"))
        self.assertFalse(self.rule_2._eval_header_value("noai"))
        self.assertFalse(self.rule_2._eval_header_value("noimageai"))
        self.assertFalse(self.rule_2._eval_header_value("other, noai"))

    def test_ai(self):
        self.assertTrue(self.rule._eval_header_value("other"))
        self.assertTrue(self.rule._eval_header_value("noindex"))
        self.assertTrue(self.rule._eval_header_value("other, noindex"))
        self.assertTrue(self.rule_2._eval_header_value("other"))
        self.assertTrue(self.rule_2._eval_header_value("noindex"))
        self.assertTrue(self.rule_2._eval_header_value("other, noindex"))

    def test_useragent_noai(self):
        self.assertFalse(self.rule._eval_header_value("spawningbot: noai"))
        self.assertFalse(self.rule._eval_header_value("spawningbot: noimageai"))
        self.assertFalse(self.rule._eval_header_value("other, spawningbot: noai"))
        self.assertFalse(self.rule._eval_header_value("other, spawningbot:noai"))
        self.assertFalse(self.rule._eval_header_value("spawningbot:other, spawningbot: noai"))
        self.assertFalse(self.rule._eval_header_value("spawningbot:other, spawningbot:noai"))
        self.assertTrue(self.rule_2._eval_header_value("spawningbot: noai"))
        self.assertTrue(self.rule_2._eval_header_value("spawningbot: noimageai"))
        self.assertTrue(self.rule_2._eval_header_value("other, spawningbot: noai"))
        self.assertTrue(self.rule_2._eval_header_value("other, spawningbot:noai"))
        self.assertTrue(self.rule_2._eval_header_value("spawningbot:other, spawningbot: noai"))
        self.assertTrue(self.rule_2._eval_header_value("spawningbot:other, spawningbot:noai"))

    def test_useragent_ai(self):
        self.assertTrue(self.rule._eval_header_value("spawningbot: all"))
        self.assertTrue(self.rule._eval_header_value("spawningbot: other"))
        self.assertTrue(self.rule._eval_header_value("other, spawningbot: all"))
        self.assertTrue(self.rule._eval_header_value("spawningbot: other, spawningbot: all, test:noai"))
        self.assertTrue(self.rule_2._eval_header_value("spawningbot: all"))
        self.assertTrue(self.rule_2._eval_header_value("spawningbot: other"))
        self.assertTrue(self.rule_2._eval_header_value("other, spawningbot: all"))
        self.assertTrue(self.rule_2._eval_header_value("spawningbot: other, spawningbot: all, test:noai"))

    def test_useragent_override(self):
        pass

    def test_stdlib(self):
        request = urllib.request.Request("http://localhost:5001/noai", data=None)
        with urllib.request.urlopen(request, timeout=3) as response:
            self.assertEqual(self.rule._handle_response(response), "noai")
            self.assertEqual(self.rule._handle_headers(response.headers), "noai")
            self.assertEqual(self.rule._handle_headers(response.getheaders()), "noai")
            self.assertFalse(self.rule.is_allowed(response=response))
            self.assertFalse(self.rule.is_allowed(headers=response.headers))

        request = urllib.request.Request("http://localhost:5001/ai", data=None)
        with urllib.request.urlopen(request, timeout=3) as response:
            self.assertEqual(self.rule._handle_response(response), "all")
            self.assertEqual(self.rule._handle_headers(response.headers), "all")
            self.assertTrue(self.rule.is_allowed(response=response))
            self.assertTrue(self.rule.is_allowed(headers=response.headers))

    def test_requests_lib(self):
        response = requests.get("http://localhost:5001/noai", timeout=3)
        self.assertEqual(self.rule._handle_response(response), "noai")
        self.assertEqual(self.rule._handle_headers(response.headers), "noai")
        self.assertFalse(self.rule.is_allowed(response=response))
        self.assertFalse(self.rule.is_allowed(headers=response.headers))

        response = requests.get("http://localhost:5001/ai")
        self.assertEqual(self.rule._handle_response(response), "all")
        self.assertEqual(self.rule._handle_headers(response.headers), "all")
        self.assertTrue(self.rule.is_allowed(response=response))
        self.assertTrue(self.rule.is_allowed(headers=response.headers))

    def test_useragent_requests(self):
        response = requests.get("http://localhost:5001/user_agents")
        self.assertTrue(self.rule.is_allowed(response=response))
        self.assertTrue(self.rule.is_allowed(headers=response.headers))

        response = requests.get("http://localhost:5001/user_agents_noai")
        self.assertFalse(self.rule.is_allowed(response=response))
        self.assertFalse(self.rule.is_allowed(headers=response.headers))

    def test_parse_useragents(self):
        response = requests.get("http://localhost:5001/user_agents")
        self.assertEqual(self.rule._handle_response(response), "demobot: noai, examplebot: noai, spawningbot: all")

    def test_malformed_headers(self):
        self.assertTrue(self.rule._eval_header_value(":,"))
        self.assertTrue(self.rule._eval_header_value(":, :, ,;: -:: "))

    def test_exceptions(self):
        self.assertRaises(dd.exceptions.XRobotsTagNoParam, self.rule.is_allowed, None, None)
        self.assertRaises(dd.exceptions.XRobotsTagUnknownHeaderObject, self.rule._handle_headers, None)
        self.assertRaises(dd.exceptions.XRobotsTagUnknownResponseObject, self.rule._handle_response, None)

    def test_url_arg(self):
        self.assertTrue(self.rule.is_allowed(url="http://localhost:5001/ai"))
        self.assertFalse(self.rule.is_allowed(url="http://localhost:5001/noai"))

    def test_noindex(self):
        rule = XRobotsTagHeader(user_agent="spawningbot", respect_noindex=False)
        self.assertTrue(rule.is_allowed(url="http://localhost:5001/noindex"))
        rule_2 = XRobotsTagHeader(user_agent="spawningbot", respect_noindex=True)
        self.assertFalse(rule_2.is_allowed(url="http://localhost:5001/noindex"))

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()
        cls.server_thread.join()



"""
This module contains the C2PA Metadata Rule class.
"""

from .base import HttpRule


# TODO: Implement, either with wrapper for CLI or python implementation of C2PA
class C2PAMetadataRule(HttpRule):
    """
    This class wraps calls to the Adobe Content Authenticity Initiative c2pa tool. TODO.
    """

    def is_allowed(self, url=None, response=None, body=None, **kwargs):
        return True

    def is_ready(self):
        return False



"""
functions to preload default evaluators and make accessible globally
"""

from .exceptions import EvaluatorAlreadyRegistered, EvaluatorNotRegistered, NotEvaluator, DefaultEvaluatorNotFound
from .evaluators import Evaluator, PreprocessEvaluator, PostprocessEvaluator

bootstrap_dictionary = {}


def load_defaults(user_agent=None):
    """Load the default evaluators."""
    register_evaluator(PreprocessEvaluator(user_agent=user_agent), overwrite=True)
    register_evaluator(PostprocessEvaluator(user_agent=user_agent), overwrite=True)


def list_evaluators():
    """List the evaluators."""
    return list(bootstrap_dictionary.keys())


def register_evaluator(evaluator, name=None, overwrite=False):
    """Register an evaluator.

    Args:
        evaluator (Evaluator): The evaluator object.
        name (str): Key name of the evaluator.
        overwrite (bool): Whether or not to overwrite the evaluator if it already exists.
    """
    if not name:
        name = evaluator.name

    if not isinstance(evaluator, Evaluator):
        raise NotEvaluator()

    if name in bootstrap_dictionary and not overwrite:
        raise EvaluatorAlreadyRegistered(name)

    bootstrap_dictionary[name] = evaluator


def deregister_evaluator(name):
    """Deregister an evaluator.

    Args:
        name (str): The name of the evaluator.
    """
    if name not in bootstrap_dictionary:
        raise EvaluatorNotRegistered(name)
    del bootstrap_dictionary[name]


def get_evaluator(name):
    """Get an evaluator.

    Args:
        name (str): The name of the evaluator.
    """
    if name not in bootstrap_dictionary:
        raise EvaluatorNotRegistered(name)
    return bootstrap_dictionary[name]


def is_allowed(name=None, **kwargs):
    """
    Check if the content is allowed.

    Args:
        name (str): The name of a specific evaluator.
        **kwargs: Arbitrary keyword arguments to read args from.
    """
    if name is not None:
        if name not in bootstrap_dictionary:
            raise EvaluatorNotRegistered(name)
        return bootstrap_dictionary[name].is_allowed(**kwargs)
    else:
        # since we are preloading evaluators manually, we can check to see which one to call
        # based on the kwargs
        if "urls" in kwargs:
            return bootstrap_dictionary["preprocess"].is_allowed(**kwargs)
        elif "url" in kwargs or "response" in kwargs or "headers" in kwargs:
            return bootstrap_dictionary["postprocess"].is_allowed(**kwargs)
        else:
            raise DefaultEvaluatorNotFound(list(kwargs.keys()))


def filter_allowed(name=None, **kwargs):
    """
    Filter a list of content.

    Args:
        name (str): The name of a specific evaluator.
        **kwargs: Arbitrary keyword arguments to read args from.
    """
    if name is not None:
        if name not in bootstrap_dictionary:
            raise EvaluatorNotRegistered(name)
        return bootstrap_dictionary[name].filter_allowed(**kwargs)
    else:
        # since we are preloading evaluators manually, we can check to see which one to call
        # based on the kwargs
        if "urls" in kwargs:
            return bootstrap_dictionary["preprocess"].filter_allowed(**kwargs)
        else:
            raise DefaultEvaluatorNotFound(list(kwargs.keys()))


load_defaults()



from datadiligence import Evaluator
from datadiligence.rules import Rule


class CustomEvaluator(Evaluator):
    def is_allowed(self, **kwargs):
        for rule in self.rules:
            if rule.is_ready() and not rule.is_allowed(**kwargs):
                return False
        return True


class CustomRule(Rule):
    def is_ready(self):
        return True

    def is_allowed(self, sample="", **kwargs):
        if "google" in sample:
            return False
        else:
            return True


class NotReadyRule(Rule):
    def is_ready(self):
        return False

    def is_allowed(self, **kwargs):
        return True


class CustomRule2(Rule):
    def is_ready(self):
        return True

    def is_allowed(self, url="", **kwargs):
        if "google" in url:
            return False
        else:
            return True




from unittest import TestCase
from datadiligence.rules import C2PAMetadataRule


class C2PATests(TestCase):

    def test_c2pa(self):
        rule = C2PAMetadataRule()
        self.assertFalse(rule.is_ready())
        self.assertTrue(rule.is_allowed(body=None))



# make sure SPAWNING_OPTS_KEY environment variable is set first

import datadiligence as dd
from pyarrow import csv
import requests

# read csv/tsv file
md = csv.read_csv("/path/to/urls.tsv", read_options=csv.ReadOptions(encoding="utf-8"), parse_options=csv.ParseOptions(delimiter='\t'))

# convert it to list
d = md.select(["url"]).to_pandas()
urls = d.iloc[:, 0].to_list()

# send through pre-process steps (only includes Spawning API currently)
verified_urls = dd.filter_allowed(urls=urls)  # use dd.is_allowed to get a list of booleans instead

# you'll probably be downloading these images in your pipeline, this is a placeholder for that
for url in verified_urls:
    response = requests.get(url)
    if response.status_code == 200:
        # this is the only new code you need here
        is_allowed = dd.is_allowed(response=response)
        if not is_allowed:
            continue

        # image is allowed after this point




from unittest import TestCase
import datadiligence as dd
from datadiligence.rules import SpawningAPI
import datadiligence.exceptions

import os
import time

# starting local server to confirm api response
from werkzeug.serving import make_server
from server.app import app
import threading
import asyncio
from contextlib import contextmanager


@contextmanager
def run_async_loop():
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        yield loop
    finally:
        loop.run_until_complete(loop.shutdown_asyncgens())
        loop.close()
        asyncio.set_event_loop(None)


def run_async_in_sync(f, *args, **kwargs):
    coro = f(*args, **kwargs)
    with run_async_loop() as loop:
        result = loop.run_until_complete(coro)
    return result


class SpawningAPITest(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.server = make_server('localhost', 5001, app)
        cls.server_thread = threading.Thread(target=cls.server.serve_forever)
        cls.server_thread.start()
        time.sleep(2)  # wait for server to start

        # set up a basic env var so most tests succeed
        if os.environ.get(SpawningAPI.API_KEY_ENV_VAR, None) is None:
            os.environ[SpawningAPI.API_KEY_ENV_VAR] = "SAMPLE_KEY"

        cls.rule = SpawningAPI(user_agent="spawningbot", chunk_size=2)
        cls.urls = [
            "https://www.spawning.ai",
            "https://www.shutterstock.com",
            "https://open.ai",
            "https://www.google.com",
            "https://laion.ai",
            "https://www.youtube.com",
        ]

    def test_chunking(self):
        iter_num = 0
        for chunk in self.rule._chunk(self.urls):
            self.assertEqual(len(chunk), 2)
            self.assertEqual(chunk[0], self.urls[iter_num * 2])
            self.assertEqual(chunk[1], self.urls[iter_num * 2 + 1])
            iter_num += 1

    def test_init(self):
        spawning_api = SpawningAPI(chunk_size=100000, timeout=-10)
        self.assertEqual(spawning_api.chunk_size, SpawningAPI.MAX_CHUNK_SIZE)
        self.assertEqual(spawning_api.timeout, SpawningAPI.DEFAULT_TIMEOUT)

    def test_is_ready(self):
        api_key = os.environ.get(SpawningAPI.API_KEY_ENV_VAR, None)
        self.assertTrue(api_key)
        tmp_spawning = SpawningAPI()
        self.assertTrue(tmp_spawning.is_ready())

        os.environ[SpawningAPI.API_KEY_ENV_VAR+"_TMP"] = api_key
        del os.environ[SpawningAPI.API_KEY_ENV_VAR]

        api_key = os.environ.get(SpawningAPI.API_KEY_ENV_VAR, None)
        self.assertFalse(api_key)
        tmp_spawning = SpawningAPI()
        self.assertFalse(tmp_spawning.is_ready())

        os.environ[SpawningAPI.API_KEY_ENV_VAR] = os.environ[SpawningAPI.API_KEY_ENV_VAR+"_TMP"]
        del os.environ[SpawningAPI.API_KEY_ENV_VAR+"_TMP"]

    def test_filter_allowed(self):
        spawning_api = SpawningAPI(max_concurrent_requests=2000)
        spawning_api.SPAWNING_AI_API_URL = "http://localhost:5001/opts"
        allowed = spawning_api.filter_allowed(urls=self.urls)
        self.assertEqual(len(allowed), 3)
        self.assertEqual(allowed[0], self.urls[1])
        self.assertEqual(allowed[1], self.urls[2])
        self.assertEqual(allowed[2], self.urls[5])

        allowed = spawning_api.filter_allowed(urls=self.urls)
        self.assertEqual(len(allowed), 3)
        self.assertEqual(allowed[0], self.urls[1])
        self.assertEqual(allowed[1], self.urls[2])
        self.assertEqual(allowed[2], self.urls[5])

        allowed = spawning_api.filter_allowed(url=self.urls[1])
        self.assertEqual(allowed[0], self.urls[1])

        with self.assertRaises(datadiligence.exceptions.SpawningNoParam):
            spawning_api.is_allowed(args=None)

    def test_is_allowed(self):
        spawning_api = SpawningAPI()
        spawning_api.SPAWNING_AI_API_URL = "http://localhost:5001/opts"
        allowed = spawning_api.is_allowed(url=self.urls[1])
        self.assertTrue(allowed)

        results = spawning_api.is_allowed(urls=self.urls)
        self.assertEqual(len(results), 6)
        self.assertFalse(results[0])
        self.assertTrue(results[1])
        self.assertTrue(results[2])
        self.assertFalse(results[3])
        self.assertFalse(results[4])
        self.assertTrue(results[5])

    def test_api_exception_handling(self):
        spawning_api = SpawningAPI(user_agent="spawningbot")
        spawning_api.SPAWNING_AI_API_URL = "http://localhost:5001/fail"
        self.assertRaises(dd.exceptions.SpawningAIAPIError, spawning_api.filter_allowed, self.urls)

        spawning_api.SPAWNING_AI_API_URL = "http://localhost:5001/fail2"
        self.assertRaises(dd.exceptions.SpawningAIAPIError, spawning_api.filter_allowed, self.urls)

        with self.assertRaises(dd.exceptions.SpawningNoParam):
            spawning_api.filter_allowed(arg=None)

    def test_api_exception_handling_async(self):
        spawning_api = SpawningAPI()
        spawning_api.SPAWNING_AI_API_URL = "http://localhost:5001/opts"
        allowed = run_async_in_sync(spawning_api.filter_allowed_async, urls=self.urls)
        self.assertEqual(len(allowed), 3)

        allowed = run_async_in_sync(spawning_api.filter_allowed_async, url=self.urls[0])
        self.assertEqual(len(allowed), 3)

        allowed = run_async_in_sync(spawning_api.is_allowed_async, urls=self.urls)
        self.assertEqual(len(allowed), 6)
        self.assertFalse(allowed[0])
        self.assertTrue(allowed[1])
        self.assertTrue(allowed[2])
        self.assertFalse(allowed[3])
        self.assertFalse(allowed[4])
        self.assertTrue(allowed[5])

        allowed = run_async_in_sync(spawning_api.is_allowed_async, url=self.urls[0])
        self.assertEqual(len(allowed), 6)

    def test_fails_async(self):
        spawning_api = SpawningAPI(user_agent="spawningbot")
        spawning_api.SPAWNING_AI_API_URL = "http://localhost:5001/fail"
        with self.assertRaises(dd.exceptions.SpawningAIAPIError):
            run_async_in_sync(spawning_api.filter_allowed_async, urls=self.urls)

        spawning_api.SPAWNING_AI_API_URL = "http://localhost:5001/fail2"
        with self.assertRaises(dd.exceptions.SpawningAIAPIError):
            run_async_in_sync(spawning_api.filter_allowed_async, urls=self.urls)

        with self.assertRaises(dd.exceptions.SpawningNoParam):
            run_async_in_sync(spawning_api.filter_allowed_async, arg=None)

        with self.assertRaises(dd.exceptions.SpawningNoParam):
            run_async_in_sync(spawning_api.is_allowed_async, arg=None)

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()
        cls.server_thread.join()



import json

from flask import Flask, Response

app = Flask(__name__)


@app.route("/ai", methods=["GET"])
def noai_headers():
    response = Response()
    response.headers["X-Robots-Tag"] = "all"
    return response


@app.route("/noai", methods=["GET"])
def ai_headers():
    response = Response()
    response.headers["X-Robots-Tag"] = "noai"
    return response


@app.route("/user_agents", methods=["GET"])
def user_agents():
    response = Response()
    response.headers["X-Robots-Tag"] = "demobot: noai, examplebot: noai, spawningbot: all"
    return response


@app.route("/user_agents_noai", methods=["GET"])
def user_agents_noai():
    response = Response()
    response.headers["X-Robots-Tag"] = "demobot: noai, examplebot: noai, spawningbot: noai"
    return response


@app.route("/noindex", methods=["GET"])
def user_agents_noindex():
    response = Response()
    response.headers["X-Robots-Tag"] = "none, noindex"
    return response


@app.route("/opts", methods=["POST"])
def opts():
    json_body = {"urls": [
        {"url": "https://www.spawning.ai", "optOut": True, "optIn": False},
        {"url": "https://www.shutterstock.com", "optOut": False, "optIn": False},
        {"url": "https://open.ai", "optOut": False, "optIn": False},
        {"url": "https://www.google.com", "optOut": True, "optIn": False},
        {"url": "https://laion.ai", "optOut": True, "optIn": True},
        {"url": "https://www.youtube.com", "optOut": False, "optIn": False}
    ]}
    return Response(json.dumps(json_body), mimetype="application/json")


@app.route("/fail", methods=["POST"])
def fail():
    response = Response()
    response.status_code = 500
    return response


@app.route("/fail2", methods=["POST"])
def fail2():
    response = Response("not-json")
    response.headers["Content-Type"] = "application/json"
    response.status_code = 200
    return response



"""
This module contains the base rule classes.
"""


class Rule:
    """
    Base class for rules. is_allowed and is_ready must be implemented.
    """

    def is_allowed(self, **kwargs):
        """Check if the request is allowed. Must be implemented.

        Args:
            **kwargs: Arbitrary keyword arguments to read args from.
        """
        raise NotImplementedError

    def is_ready(self):
        """Check if the rule is ready to be used."""
        raise NotImplementedError


class BulkRule(Rule):
    """
    Base class for bulk rules. filter_allowed and is_ready must be implemented.
    """

    def filter_allowed(self, **kwargs):
        """Filter a list of entries based on the rules in this evaluator."""
        raise NotImplementedError


class HttpRule(Rule):
    def __init__(self, user_agent=None):
        """Initialize the rule with a user agent.

        Args:
            user_agent (str): The user agent to pass on to the rules.
        """
        super().__init__()
        self.user_agent = user_agent




from unittest import TestCase
from datadiligence.utils import get_url
import requests


class TestUtils(TestCase):

    def test_get_url(self):
        response = get_url("http://www.google.com")
        self.assertTrue(response.ok)

        self.assertRaises(Exception, get_url, "wesvadsvrwabg")

        user_agent = requests.utils.default_user_agent()
        response = get_url("http://www.google.com", user_agent=user_agent)
        self.assertTrue(response.ok)



# make sure SPAWNING_OPTS_KEY environment variable is set first

import datadiligence as dd
from pyarrow import csv

# read csv/tsv file
md = csv.read_csv("~/Downloads/cc40k.tsv", read_options=csv.ReadOptions(encoding="utf-8"), parse_options=csv.ParseOptions(delimiter='\t'))

# convert it to list
d = md.select(["url"]).to_pandas()
urls = d.iloc[:, 0].to_list()

# send through pre-process steps (only includes Spawning API currently)
verified_urls = dd.filter_allowed(urls=urls)



"""Postprocess evaluator module."""

from .base import Evaluator
from ..rules import XRobotsTagHeader


class PostprocessEvaluator(Evaluator):
    """
    Postprocess Evaluator class. Loads XRobotsTagHeader rule by default.
    """
    name = "postprocess"

    def __init__(self, user_agent=None):
        super().__init__()
        self.add_rule(XRobotsTagHeader(user_agent))

    def is_allowed(self, **kwargs):
        """Check if the headers are allowed based on the rules in this evaluator.

        Args:
            **response (http.client.HTTPResponse|requests.Response): The response object.
            **headers (dict|http.client.HTTPMessage): The headers dictionary.

        Returns:
            bool: True if the content is allowed, False otherwise.
        """
        for rule in self.rules:
            if rule.is_ready() and not rule.is_allowed(**kwargs):
                return False
        return True



"""
This module wraps HTTP calls to the Spawning AI API.
See Spawning API documentation here: https://opts-api.spawningaiapi.com/docs.

"""

import requests
import requests.utils
from requests.adapters import HTTPAdapter, Retry
import aiohttp
import asyncio
import os
from ..exceptions import SpawningAIAPIError, SpawningNoParam
from .base import BulkRule
from concurrent.futures import ThreadPoolExecutor, wait
import itertools
import sys


class SpawningAPI(BulkRule):
    """
    This class wraps basic requests to the Spawning API.
    """

    API_KEY_ENV_VAR = "SPAWNING_OPTS_KEY"
    MAX_CHUNK_SIZE = 10000
    SPAWNING_AI_API_URL = "https://opts-api.spawningaiapi.com/api/v2/query/urls/"
    DEFAULT_TIMEOUT = 20
    MAX_CONCURRENT_REQUESTS = 10

    def __init__(self, user_agent=None, chunk_size=MAX_CHUNK_SIZE, timeout=DEFAULT_TIMEOUT, max_retries=5,
                 max_concurrent_requests=MAX_CONCURRENT_REQUESTS):
        """Create a new SpawningAPI BulkRole instance.

        Args:
            user_agent (str): The user agent to use when making requests to the Spawning AI API.
            chunk_size (int): The maximum number of URLs to submit in a single request.
            timeout (int): The maximum number of seconds to wait for a response from the Spawning AI API.
            max_retries (int): The maximum number of times to retry a request to the Spawning AI API.
            max_concurrent_requests (int): The maximum number of concurrent requests to the Spawning AI API.
        """

        super().__init__()

        # check if API key is installed
        self.api_key = os.environ.get(SpawningAPI.API_KEY_ENV_VAR, "")
        if not self.api_key:
            print(
                "Spawning API key not found. Please set your API key to the environment variable "
                + SpawningAPI.API_KEY_ENV_VAR
            )

        if user_agent is None:
            self.user_agent = requests.utils.default_user_agent()
        else:
            self.user_agent = user_agent

        self.max_retries = max_retries

        if chunk_size > self.MAX_CHUNK_SIZE or chunk_size is None:
            chunk_size = self.MAX_CHUNK_SIZE
        self.chunk_size = chunk_size

        if timeout < 0 or timeout is None:
            timeout = self.DEFAULT_TIMEOUT
        self.timeout = timeout

        if max_concurrent_requests <= 0 \
                or max_concurrent_requests > self.MAX_CONCURRENT_REQUESTS \
                or max_concurrent_requests is None:
            max_concurrent_requests = self.MAX_CONCURRENT_REQUESTS
        self.max_concurrent_requests = max_concurrent_requests

    def filter_allowed(self, urls=None, url=None, **kwargs):
        """Submit a list of URLs to the Spawning AI API.
        Args:
            urls (list): A list of URLs to submit.
            url (str): A single URL to submit.
        Returns:
            list: A list containing the allowed urls of the submission.
        """
        if urls is None:
            if url is not None:
                urls = [url]
            else:
                raise SpawningNoParam()

        results = []

        # thread pool executor to submit chunks in parallel
        with ThreadPoolExecutor(max_workers=self.max_concurrent_requests) as executor:
            futures = [executor.submit(self._submit_chunk, chunk) for chunk in self._chunk(urls)]
            wait(futures)

            for future in futures:
                results.extend([response_url["url"] for response_url in future.result() if not response_url["optOut"]])

        return results

    def is_allowed(self, urls=None, url=None, **kwargs):
        """Submit a list of URLs to the Spawning AI API.
        Args:
            urls (list): A list of URLs to submit.
            url (str): A single URL to submit.
        Returns:
            list: A list containing booleans, indicating if a respective URL is allowed.
        """
        if urls is None:
            if url is not None:
                urls = [url]
            else:
                raise SpawningNoParam()

        results = []

        # thread pool executor to submit chunks in parallel
        with ThreadPoolExecutor(max_workers=self.max_concurrent_requests) as executor:
            futures = [executor.submit(self._submit_chunk, chunk) for chunk in self._chunk(urls)]
            wait(futures)

            for future in futures:
                results.extend([not response_url["optOut"] for response_url in future.result()])

        return results

    async def filter_allowed_async(self, urls=None, url=None, **kwargs):
        """Submit a list of URLs to the Spawning AI API.

        Args:
            urls (list): A list of URLs to submit.
            url (str): A single URL to submit.

        Returns:
            list: A list containing the allowed URLs.

        """
        if urls is None:
            if url is not None:
                urls = [url]
            else:
                raise SpawningNoParam()

        results = await self._submit_chunks_async(urls)
        results = [response_url["url"] for response_url in results if not response_url["optOut"]]
        return results

    async def is_allowed_async(self, urls=None, url=None, **kwargs):
        """Submit a list of URLs to the Spawning AI API.

        Args:
            urls (list): A list of URLs to submit.
            url (str): A single URL to submit.

        Returns:
            list: A list of boolean values indicating if a URL is allowed to be used or not.

        """
        if urls is None:
            if url is not None:
                urls = [url]
            else:
                raise SpawningNoParam()

        results = await self._submit_chunks_async(urls)
        results = [not response_url["optOut"] for response_url in results]
        return results

    def is_ready(self):
        """Check if the Spawning AI API is ready to be used. This is determined by whether or not the API key is set.
        Returns:
            bool: True if the API key is set, False otherwise.
        """
        return bool(self.api_key)

    def _chunk(self, urls):
        """Generator to split a list of URLs into chunks.

        Args:
            urls (list): A list of URLs to split into chunks.

        Yields:
            list: A list of URLs.

        """

        for i in range(0, len(urls), self.chunk_size):
            yield urls[i:i + self.chunk_size]

    async def _submit_chunks_async(self, urls):
        """Submit a list of URLs to the Spawning AI API.

        Args:
            urls (list): A list of URLs to submit.

        Returns:
            list: A list containing the result dicts of the submission.

        """

        # build out http tasks
        tasks = []
        headers = {
            "User-Agent": self.user_agent,
            "Content-Type": "text/plain",
            "Authorization": "API " + self.api_key
        }
        semaphore = asyncio.Semaphore(self.max_concurrent_requests)
        async with aiohttp.ClientSession(headers=headers) as session:
            # if 3.6 or earlier, run create_task from loop
            if sys.version_info < (3, 7):
                loop = asyncio.get_event_loop()
                for chunk in self._chunk(urls):
                    tasks.append(loop.create_task(self._submit_chunk_async(session, semaphore, chunk)))
            else:
                for chunk in self._chunk(urls):
                    tasks.append(asyncio.create_task(self._submit_chunk_async(session, semaphore, chunk)))

            results = await asyncio.gather(*tasks)

        # flatten
        return [response_url for response_url in itertools.chain(*results)]

    async def _submit_chunk_async(self, session, semaphore, urls):
        """Submit a chunk of URLs to the Spawning AI API asynchronously.

        Args:
            session (aiohttp.ClientSession): The aiohttp session to use.
            urls (list): A list of URLs to submit.

        Returns:
            list: A list containing the result dicts of the submission.

        """

        # limit concurrent requests
        async with semaphore:
            # create body
            body = "\n".join(urls)

            # make request
            async with session.post(self.SPAWNING_AI_API_URL, data=body) as response:
                try:
                    if response.status != 200:
                        raise SpawningAIAPIError("Spawning AI API returned a non-200 status code: " +
                                                 str(response.status))

                    results = await response.json()
                    return results.get("urls", [])
                except Exception as e:
                    raise SpawningAIAPIError(e)

    def _submit_chunk(self, urls):
        """Submit a chunk of URLs to the Spawning AI API.
        Args:
            urls (list): A list of URLs to submit.
        Returns:
            list: A list containing the result dicts of the submission.
        """

        # create body
        body = "\n".join(urls)

        # make request
        try:
            s = requests.Session()
            retries = Retry(total=self.max_retries, backoff_factor=0.1, status_forcelist=[429, 500, 502, 503, 504])
            s.mount("https://", HTTPAdapter(max_retries=retries))
            response = s.post(self.SPAWNING_AI_API_URL,
                              data=body,
                              headers={
                                  "User-Agent": self.user_agent,
                                  "Content-Type": "text/plain",
                                  "Authorization": "API " + self.api_key
                              },
                              timeout=self.timeout
                              )
            if response.status_code != 200:
                raise SpawningAIAPIError("Spawning AI API returned a non-200 status code: " + str(response.status_code))

            return response.json().get("urls", [])
        except Exception as e:
            raise SpawningAIAPIError(e)



"""
Utility functions for package.
"""
import requests


def get_url(url, user_agent=None):
    """
    Get the URL and return the response object.

    Args:
        url (str): The URL to get.
        user_agent (str): The user agent to use.

    Returns:
        requests.Response: The response object.
    """

    # TODO: add a cache so requests aren't made twice for the same URL

    if not user_agent:
        user_agent = requests.utils.default_user_agent()

    session = requests.Session()
    return session.get(url, headers={"User-Agent": user_agent}, timeout=10)



"""
Rules to manage validation using HTTP properties
"""

import http.client
import requests
from ..exceptions import XRobotsTagNoParam, XRobotsTagUnknownHeaderObject, XRobotsTagUnknownResponseObject
from ..utils import get_url
from .base import HttpRule


class XRobotsTagHeader(HttpRule):
    """
    This class wraps logic to read the X-Robots-Tag header.
    """
    AI_DISALLOWED_VALUES = ["noai", "noimageai"]
    INDEX_DISALLOWED_VALUES = ["noindex", "none", "noimageindex", "noai", "noimageai"]
    HEADER_NAME = "X-Robots-Tag"

    def __init__(self, user_agent=None, respect_noindex=False):
        """Create a new XRobotsTagHeader instance.

        Args:
            user_agent (str): The user agent to use when making requests to the Spawning AI API.
            respect_noindex (bool): If True, index rules will be respected alongside AI rules.
        """
        super().__init__(user_agent=user_agent)

        # index rules aren't for AI, so we ignore them by default.
        # They could have been delivered/found by any number of other means, even for internal use
        if respect_noindex:
            self.disallowed_headers = self.INDEX_DISALLOWED_VALUES
        else:
            self.disallowed_headers = self.AI_DISALLOWED_VALUES

    def is_allowed(self, url=None, response=None, headers=None, **kwargs):
        """Check if the X-Robots-Tag header allows the user agent to access the resource.

        Args:
            url: (str): The URL of the resource.
            response (http.client.HTTPResponse|requests.Response, optional): The response object. Defaults to None
            headers (dict|http.client.HTTPMessage, optional): The headers dictionary. Defaults to None.

        Returns:
            bool: True if the user agent is allowed to access the resource, False otherwise.
        """

        if headers:
            header_value = self._handle_headers(headers)
        elif response:
            header_value = self._handle_response(response)
        elif url:
            response = self._handle_url(url)
            header_value = self._handle_headers(response.headers)
        else:
            raise XRobotsTagNoParam()

        return self._eval_header_value(header_value, **kwargs)

    def is_ready(self):
        """
        This rule is always ready.
        """
        return True

    def _eval_header_value(self, header_value, user_agent=None, **kwargs):
        """
        Evaluate the header value to determine if the user agent is allowed to access the resource.

        Args:
            header_value (str): The header value.
            user_agent (str): Override user agent to use when making requests to the Spawning AI API.

        Returns:
            bool: True if the user agent is allowed to access the resource, False otherwise.
        """
        if not header_value:
            return True

        # if we have a specific user agent
        if not user_agent:
            user_agent = self.user_agent

        # check if blocking all user agents
        for value in header_value.split(","):
            if value.strip() in self.disallowed_headers:
                return False

            # check if blocking specific user agent
            if user_agent:
                ua_values = value.split(":")
                if len(ua_values) == 2 and ua_values[0].strip() == user_agent \
                        and ua_values[1].strip() in self.disallowed_headers:
                    return False

        return True

    def _handle_response(self, response):
        """
        Handle the response object to get the header value.

        Args:
            response (http.client.HTTPResponse|requests.Response): The response object.

        Returns:
            str: The header value.
        """
        if type(response) == http.client.HTTPResponse:
            header_value = response.getheader(XRobotsTagHeader.HEADER_NAME, "")
        elif type(response) == requests.Response:
            header_value = response.headers.get(XRobotsTagHeader.HEADER_NAME, "")
        else:
            raise XRobotsTagUnknownResponseObject()
        return header_value

    def _handle_headers(self, headers):
        """
        Handle the headers object to get the header value.

        Args:
            headers (dict|http.client.HTTPMessage|CaseInsensitiveDict): The headers object.

        Returns:
            str: The header value.
        """
        if type(headers) == dict or type(headers) == requests.structures.CaseInsensitiveDict:
            header_value = headers.get(XRobotsTagHeader.HEADER_NAME, "")
        elif type(headers) == list and len(headers) > 0 and type(headers[0]) == tuple:
            header_value = dict(headers).get(XRobotsTagHeader.HEADER_NAME, "")
        elif type(headers) == http.client.HTTPMessage:
            header_value = headers.get(XRobotsTagHeader.HEADER_NAME, "")
        else:
            raise XRobotsTagUnknownHeaderObject()

        return header_value

    def _handle_url(self, url):
        """
        If given a raw URL, submit a request to get the image.

        Args:
            url (str): The URL of the resource.

        Returns:
            requests.Response: Response object.
        """
        return get_url(url, user_agent=self.user_agent)



"""
This module contains the base Evaluator class.
"""
from ..rules import Rule


class Evaluator:
    """
    Base class for evaluators. is_allowed must be implemented.
    """
    name = "base_evaluator"

    def __init__(self):
        self.rules = []

    def add_rule(self, rule):
        """Add a rule to the evaluator."""
        if isinstance(rule, Rule):
            self.rules.append(rule)

    def is_allowed(self, **kwargs):
        """Check each rule to see if the request is allowed.

        Args:
            **kwargs (any): Keyword args to pass to rule

        Returns:
            bool: True if the content is allowed, False otherwise.
        """
        for rule in self.rules:
            if rule.is_ready() and not rule.is_allowed(**kwargs):
                return False
        return True

    def filter_allowed(self, **kwargs):
        """Filter a list of entries based on the rules in this evaluator."""
        raise NotImplementedError



"""
This module contains the HttpEvaluator class.
"""

from .base import Evaluator
from ..rules import XRobotsTagHeader


class HttpEvaluator(Evaluator):
    """
    HTTP Evaluator class. Loads XRobotsTagHeader rule by default.
    """
    name = "http"

    def __init__(self, user_agent=None, respect_robots=True):
        """Load the default rules.

        Args:
            user_agent (str): The user agent to pass on to the rules.
            respect_robots (bool): Whether to respect the X-Robots-Tag header.
        """
        super().__init__()
        if respect_robots:
            self.rules.append(XRobotsTagHeader(user_agent))



"""
Exceptions for the package.
"""


class XRobotsTagNoParam(Exception):
    """
    Raised when XRobotsTagHeader isn't provided with either an url, response, or headers object.
    """

    def __init__(self):
        super().__init__("XRobotsTagHeader must be provided with either an url, response, or headers object.")


class XRobotsTagUnknownHeaderObject(Exception):
    """
    Raised when XRobotsTagHeader is provided with an unknown header object.
    """

    def __init__(self):
        super().__init__(
            "XRobotsTagHeader must be provided with a header object of types "
            "dict|CaseInsensitiveDict|http.client.HTTPMessage.")


class XRobotsTagUnknownResponseObject(Exception):
    """
    Raised when XRobotsTagHeader is provided with an unknown response object.
    """

    def __init__(self):
        super().__init__(
            "XRobotsTagHeader must be provided with a response object of types "
            "http.client.HTTPResponse|requests.Response.")


class SpawningAIAPIError(Exception):
    """
    Raised when the Spawning AI API returns an error.
    """

    def __init__(self, message):
        super().__init__(message)


class EvaluatorNotRegistered(Exception):
    """
    Raised when an evaluator is not registered.
    """

    def __init__(self, name):
        super().__init__(f"Evaluator {name} not registered.")


class DefaultEvaluatorNotFound(Exception):
    """
    Raised when aa default evaluator can't be determined.
    """

    def __init__(self, args):
        super().__init__(f"No default evaluator found which can handle the following arguments {args}.")


class NotEvaluator(Exception):
    """
    Raised when an object is not an evaluator.
    """

    def __init__(self):
        super().__init__("Object must be of type Evaluator.")


class EvaluatorAlreadyRegistered(Exception):
    """
    Raised when an evaluator is already registered.
    """

    def __init__(self, name):
        super().__init__(f"Evaluator {name} already registered.")


class SpawningNoParam(Exception):
    """
    Raised when SpawningAPI isn't provided with a list of urls.
    """

    def __init__(self):
        super().__init__("SpawningAPI must be provided with a list of urls.")



"""
This module contains default Evaluators.
"""

from .base import Evaluator
from .http import HttpEvaluator
from .postprocess import PostprocessEvaluator
from .preprocess import PreprocessEvaluator




from unittest import TestCase
import datadiligence as dd
import datadiligence.exceptions
from samples.custom import CustomEvaluator, CustomRule
import urllib
import time

# starting local server to confirm api response
from werkzeug.serving import make_server
from server.app import app
import threading
import os
from datadiligence.rules import SpawningAPI


class BootstrapTests(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.server = make_server('localhost', 5001, app)
        cls.server_thread = threading.Thread(target=cls.server.serve_forever)
        cls.server_thread.start()
        time.sleep(2)

        # set up a basic env var so most tests succeed
        if os.environ.get(SpawningAPI.API_KEY_ENV_VAR, None) is None:
            os.environ[SpawningAPI.API_KEY_ENV_VAR] = "SAMPLE_KEY"

        cls.urls = [
            "https://www.spawning.ai",
            "https://www.shutterstock.com",
            "https://open.ai",
            "https://www.google.com",
            "https://laion.ai",
            "https://www.youtube.com",
        ]

    def test_exceptions(self):
        self.assertRaises(dd.exceptions.EvaluatorNotRegistered, dd.is_allowed, "x")
        self.assertRaises(dd.exceptions.NotEvaluator, dd.register_evaluator, None, "y")

        custom_evaluator = dd.Evaluator()
        dd.register_evaluator(custom_evaluator, "custom0")
        self.assertRaises(dd.exceptions.EvaluatorAlreadyRegistered,
                          dd.register_evaluator, custom_evaluator, "custom0")
        self.assertRaises(dd.exceptions.EvaluatorNotRegistered, dd.deregister_evaluator, "not-used")
        self.assertRaises(dd.exceptions.EvaluatorNotRegistered, dd.get_evaluator, "not-used")
        self.assertRaises(dd.exceptions.DefaultEvaluatorNotFound, dd.is_allowed, example="not-used")

        with self.assertRaises(datadiligence.exceptions.EvaluatorNotRegistered):
            dd.filter_allowed(name="none", urls=[])

        with self.assertRaises(datadiligence.exceptions.DefaultEvaluatorNotFound):
            dd.filter_allowed(args=None)

    def test_load_defaults(self):
        # reset evaluators
        for key in dd.list_evaluators():
            dd.deregister_evaluator(key)

        dd.load_defaults()
        self.assertTrue(isinstance(dd.get_evaluator("preprocess"), dd.PreprocessEvaluator))
        self.assertTrue(isinstance(dd.get_evaluator("postprocess"), dd.PostprocessEvaluator))
        self.assertEqual(len(dd.list_evaluators()), 2)

    def test_register_evaluator(self):
        custom_evaluator = dd.Evaluator()
        dd.register_evaluator(custom_evaluator, "custom1")
        self.assertEqual(dd.get_evaluator("custom1"), custom_evaluator)

    def test_deregister_evaluator(self):
        custom_evaluator = dd.Evaluator()
        dd.register_evaluator(custom_evaluator, "custom3")
        self.assertEqual(dd.get_evaluator("custom3"), custom_evaluator)
        dd.deregister_evaluator("custom3")
        self.assertRaises(dd.exceptions.EvaluatorNotRegistered, dd.deregister_evaluator, "custom3")

    def test_is_allowed(self):
        custom_evaluator = CustomEvaluator()
        custom_evaluator.add_rule(CustomRule())
        dd.register_evaluator(custom_evaluator, "custom2")
        self.assertFalse(dd.is_allowed("custom2", sample="www.google.com"))
        self.assertTrue(dd.is_allowed("custom2", sample="www.example.com"))

    def test_filter_allowed(self):

        dd.load_defaults()

        request = urllib.request.Request("http://localhost:5001/noai", data=None)
        with urllib.request.urlopen(request, timeout=3) as response:
            self.assertFalse(dd.is_allowed(response=response))

        # hack to reach local instance
        dd.get_evaluator("preprocess").rules[0].SPAWNING_AI_API_URL = "http://localhost:5001/opts"
        filtered_urls = dd.filter_allowed(urls=self.urls)
        self.assertEqual(len(filtered_urls), 3)
        self.assertEqual(filtered_urls[0], self.urls[1])
        self.assertEqual(filtered_urls[1], self.urls[2])
        self.assertEqual(filtered_urls[2], self.urls[5])

        dd.load_defaults()

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()
        cls.server_thread.join()



"""
This module contains the PreprocessEvaluator class.
"""

from .base import Evaluator
from ..rules import SpawningAPI, BulkRule


class PreprocessEvaluator(Evaluator):
    """
    Preprocess Evaluator class. Loads SpawningAPI rule by default.
    """
    name = "preprocess"

    def __init__(self, user_agent=None):
        """ Load the default rules.

        Args:
            user_agent (str): The user agent to pass on to the rules.

        """
        super().__init__()
        self.add_rule(SpawningAPI(user_agent))

    def add_rule(self, rule):
        """Add a rule to the evaluator."""
        if issubclass(rule.__class__, BulkRule):
            self.rules.append(rule)

    def filter_allowed(self, urls=None, **kwargs):
        """Filter a list of urls based on the rules in this evaluator.

        Args:
            urls (list): A list of urls to filter.
            **kwargs: Arbitrary keyword arguments to read args from.

        Returns:
            list: A list of urls that are allowed.
        """

        if urls is None:
            return []

        allowed = urls
        for rule in self.rules:
            # if everything is already filtered out, stop
            if len(allowed) == 0:
                break
            if rule.is_ready():
                allowed = rule.filter_allowed(urls=allowed)

        return allowed

    def is_allowed(self, urls=None, *kwargs):
        """
        Check if the urls are allowed.

        Args:
            urls (list): A list of urls to check.
            **kwargs: Arbitrary keyword arguments to read args from.

        Returns:
            bool: List of boolean values, respectively indicating if can be used or not
        """

        if urls is None:
            return []

        allowed = [True] * len(urls)
        for rule in self.rules:
            if rule.is_ready():
                rule_results = rule.is_allowed(urls=urls)

                # update allowed list to False only if rule_results is False
                allowed = [a and b for a, b in zip(allowed, rule_results)]

        return allowed



import requests
import urllib.request
from unittest import TestCase
import datadiligence as dd
from datadiligence import HttpEvaluator, PreprocessEvaluator, PostprocessEvaluator
from datadiligence.rules import SpawningAPI, XRobotsTagHeader
import time
import os
from samples.custom import CustomEvaluator, NotReadyRule, CustomRule2

# starting local server to echo back headers
from werkzeug.serving import make_server
from server.app import app
import threading
import datadiligence.exceptions


class EvaluatorTests(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.server = make_server('localhost', 5001, app)
        cls.server_thread = threading.Thread(target=cls.server.serve_forever)
        cls.server_thread.start()
        time.sleep(1)  # wait for server to start

    def test_http_evaluator(self):
        http_evaluator = HttpEvaluator()
        self.assertTrue(len(http_evaluator.rules) > 0)

        response = requests.get("http://localhost:5001/noai")
        self.assertFalse(http_evaluator.is_allowed(response=response))
        self.assertFalse(http_evaluator.is_allowed(headers=response.headers))

        request = urllib.request.Request("http://localhost:5001/noai", data=None)
        with urllib.request.urlopen(request, timeout=3) as response:
            self.assertFalse(http_evaluator.is_allowed(response=response))
            self.assertFalse(http_evaluator.is_allowed(headers=response.headers))

        response = requests.get("http://localhost:5001/ai")
        self.assertTrue(http_evaluator.is_allowed(response=response))
        self.assertTrue(http_evaluator.is_allowed(headers=response.headers))

        http_evaluator_2 = HttpEvaluator(respect_robots=False)
        self.assertEqual(len(http_evaluator_2.rules), 0)

    def test_custom_evaluator(self):
        # custom evaluator
        custom_evaluator = CustomEvaluator()
        custom_rule = CustomRule2()
        not_ready_rule = NotReadyRule()
        custom_evaluator.add_rule(custom_rule)
        custom_evaluator.add_rule(not_ready_rule)
        self.assertFalse(custom_evaluator.is_allowed(url="https://www.google.com"))
        self.assertTrue(custom_evaluator.is_allowed(url="https://www.spawning.ai"))

        len_rules = len(custom_evaluator.rules)
        custom_evaluator.add_rule(None)
        self.assertEqual(len(custom_evaluator.rules), len_rules)

    def test_bulk_evaluator(self):
        bulk_evaluator = PreprocessEvaluator()
        bulk_evaluator.rules = []

        # patch to point to local dev
        if os.environ.get(SpawningAPI.API_KEY_ENV_VAR, None) is None:
            os.environ[SpawningAPI.API_KEY_ENV_VAR] = "SAMPLE_KEY"
        spawning_api = SpawningAPI(user_agent="spawningbot")
        spawning_api.SPAWNING_AI_API_URL = "http://localhost:5001/opts"

        self.assertEqual(len(bulk_evaluator.rules), 0)
        bulk_evaluator.add_rule(spawning_api)
        self.assertEqual(len(bulk_evaluator.rules), 1)

        urls = bulk_evaluator.filter_allowed([
            "https://www.spawning.ai",
            "https://www.shutterstock.com",
            "https://open.ai",
            "https://www.google.com",
            "https://laion.ai",
            "https://www.youtube.com",
        ])
        self.assertEqual(len(urls), 3)

        urls = bulk_evaluator.filter_allowed(urls=[
            "https://www.spawning.ai",
            "https://www.shutterstock.com",
            "https://open.ai",
            "https://www.google.com",
            "https://laion.ai",
            "https://www.youtube.com",
        ])
        self.assertEqual(len(urls), 3)

        self.assertEqual(len(bulk_evaluator.rules), 1)
        bulk_evaluator.add_rule(None)
        self.assertEqual(len(bulk_evaluator.rules), 1)

        bulk_evaluator = PreprocessEvaluator()
        self.assertTrue(len(bulk_evaluator.rules) > 0)

        self.assertEqual(bulk_evaluator.filter_allowed([]), [])
        self.assertEqual(bulk_evaluator.filter_allowed(None), [])

        self.assertEqual(bulk_evaluator.is_allowed(), [])

    def test_bulk_evaluator_ready_state(self):
        b = PreprocessEvaluator()
        b.rules = []
        del os.environ[SpawningAPI.API_KEY_ENV_VAR]
        spawning_rule = SpawningAPI()

        self.assertFalse(spawning_rule.is_ready())

        b.add_rule(spawning_rule)

        urls = b.filter_allowed([
            "https://www.spawning.ai",
            "https://www.shutterstock.com",
            "https://open.ai",
            "https://www.google.com",
            "https://laion.ai",
            "https://www.youtube.com",
        ])
        self.assertEqual(len(urls), 6)

    def test_postprocessor(self):
        postprocessor = PostprocessEvaluator()
        postprocessor.rules = []
        not_ready_rule = NotReadyRule()
        custom_rule = CustomRule2()
        postprocessor.add_rule(not_ready_rule)
        postprocessor.add_rule(custom_rule)

        # custom rule is saying it's not ready, so we don't use it to evaluate
        self.assertTrue(postprocessor.is_allowed(url="http://localhost:5001/noai"))

        # but later rules still work correctly
        self.assertFalse(postprocessor.is_allowed(url="http://google.com"))

    def test_default_eval_args(self):
        """
        Testing named variables to root package evaluator routing
        """

        # http defaults
        self.assertFalse(dd.is_allowed(url="http://localhost:5001/noai"))

        # patch to point to local dev
        if os.environ.get(SpawningAPI.API_KEY_ENV_VAR, None) is None:
            os.environ[SpawningAPI.API_KEY_ENV_VAR] = "SAMPLE_KEY"
        spawning_api = SpawningAPI(user_agent="spawningbot")
        spawning_api.SPAWNING_AI_API_URL = "http://localhost:5001/opts"
        evaluator = dd.get_evaluator("preprocess")
        evaluator.rules = []
        evaluator.add_rule(XRobotsTagHeader())
        evaluator.add_rule(spawning_api)
        dd.register_evaluator(evaluator, name="preprocess", overwrite=True)

        # urls defaults
        urls = dd.filter_allowed(urls=[
            "https://www.spawning.ai",
            "https://www.shutterstock.com",
            "https://open.ai",
            "https://www.google.com",
            "https://laion.ai",
            "https://www.youtube.com",
        ])
        self.assertEqual(len(urls), 3)

        bool_urls = dd.is_allowed(urls=[
            "https://www.spawning.ai",
            "https://www.shutterstock.com",
            "https://open.ai",
            "https://www.google.com",
            "https://laion.ai",
            "https://www.youtube.com",
        ])
        self.assertFalse(bool_urls[0])
        self.assertTrue(bool_urls[1])
        self.assertTrue(bool_urls[2])
        self.assertFalse(bool_urls[3])
        self.assertFalse(bool_urls[4])
        self.assertTrue(bool_urls[5])

        # response and headers defaults
        response = requests.get("http://localhost:5001/noai")
        self.assertFalse(dd.is_allowed(response=response))
        self.assertFalse(dd.is_allowed(headers=response.headers))

        response = requests.get("http://localhost:5001/ai")
        self.assertTrue(dd.is_allowed(response=response))
        self.assertTrue(dd.is_allowed(headers=response.headers))

        urls = dd.filter_allowed(name="preprocess", urls=[
            "https://www.spawning.ai",
            "https://www.shutterstock.com",
            "https://open.ai",
            "https://www.google.com",
            "https://laion.ai",
            "https://www.youtube.com",
        ])
        self.assertEqual(len(urls), 3)

        # reload standard evaluators
        dd.load_defaults()

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()
        cls.server_thread.join()
